#ifndef COCO_GMOCK_HELPERS_H_
#define COCO_GMOCK_HELPERS_H_

// MIT License
//
// Copyright (c) 2021 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include <functional>
#include <map>
#include <memory>
#include <mutex>
#include <queue>
#include <thread>

#include "coco/runtime.h"
#include "gmock/gmock.h"

#if !defined(__clang__) && defined(__GNUC__) && (__GNUC__ < 4 || (__GNUC__ == 4 && __GNUC_MINOR__ < 5))
#define COCO_SLEEP_WORKAROUND
#include <ctime>
#endif

namespace coco {

/// A simple helper class to assist with creating multi-threaded gmocks.
///
/// This uses a background thread to execute actions asynchronously. The intended use is that you create a single
/// instance of MultiThreadedGMockHelper inside your test case, and use InvokeAsync to invoke actions asynchronously.
/// These will then be executed by a background thread. For example:
///
///     ClientMock client_mock;
///     coco::MultiThreadedGMockHelper helper;
///     ...
///     EXPECT_CALL(client_mock, pinger_pong(value))
///       .WillOnce(helper.InvokeAsync(Invoke(&client_mock.pinger().connected(), &PPingerProvided::terminate)))
class MultiThreadedGMockHelper {
 private:
  class ExecuteAsyncAction;

 public:
  MultiThreadedGMockHelper() : shutdown_(false), thread_(std::bind(&MultiThreadedGMockHelper::run, this)) {}

  ~MultiThreadedGMockHelper() {
    {
      std::lock_guard<std::mutex> lock(mutex_);
      shutdown_ = true;
      thread_condition_.notify_one();
    }
    thread_.join();
  }

  /// Executes an arbitrary gMock action asynchronously.
  ///
  /// Note that this will be immediately added to a queue which will execute concurrently with this, meaning that the
  /// execution of the action could happen before this function even returns.
  testing::PolymorphicAction<ExecuteAsyncAction> InvokeAsync(const testing::Action<void()>& action) {
    return testing::MakePolymorphicAction(ExecuteAsyncAction(*this, action, CocoClock::duration::zero()));
  }

  /// Executes an arbitrary gMock action asynchronously after the specified delay has elapsed.
  testing::PolymorphicAction<ExecuteAsyncAction> InvokeAsync(const testing::Action<void()>& action,
                                                             CocoClock::duration delay) {
    return testing::MakePolymorphicAction(ExecuteAsyncAction(*this, action, delay));
  }

  void execute_async(std::function<void()>&& function) {
    std::lock_guard<std::mutex> lock(mutex_);
    functions_.push(std::move(function));
    thread_condition_.notify_one();
  }

  void execute_async(std::function<void()>&& function, CocoClock::duration delay) {
    std::lock_guard<std::mutex> lock(mutex_);
    unsigned int timer_id;
    if (free_timer_ids_.empty()) {
      timer_id = timer_functions_.size();
    } else {
      timer_id = free_timer_ids_.back();
      free_timer_ids_.pop_back();
    }
    timer_queue_.enable_timer(timer_id, delay);
    timer_functions_.insert({timer_id, std::move(function)});
    thread_condition_.notify_one();
  }

 private:
  class ExecuteAsyncAction {
   public:
    ExecuteAsyncAction(MultiThreadedGMockHelper& executor, const testing::Action<void()>& action,
                       CocoClock::duration delay)
        : executor(executor), action(action), delay(delay) {}

    template <typename Result, typename ArgumentTuple>
    Result Perform(const ArgumentTuple& args) {
      if (delay == CocoClock::duration::zero()) {
        executor.execute_async(Executor(action));
      } else {
        executor.execute_async(Executor(action), delay);
      }
    }

   private:
    MultiThreadedGMockHelper& executor;
    testing::Action<void()> action;
    CocoClock::duration delay;

    class Executor {
     public:
      Executor(const testing::Action<void()>& action) : action(action) {}
      void operator()() { action.Perform(testing::Action<void()>::ArgumentTuple()); }

     private:
      testing::Action<void()> action;
    };
  };

  bool shutdown_;
  std::condition_variable thread_condition_;
  std::queue<std::function<void()>> functions_;
  std::map<unsigned int, std::function<void()>> timer_functions_;
  std::vector<unsigned int> free_timer_ids_;
  VariableSizeTimerQueue<unsigned int> timer_queue_;
  std::mutex mutex_;
  std::thread thread_;

  class Unlock {
   public:
    Unlock(std::unique_lock<std::mutex>& lock) : lock(lock) { lock.unlock(); }
    ~Unlock() { lock.lock(); }

   private:
    std::unique_lock<std::mutex>& lock;
  };

  void run() {
    std::unique_lock<std::mutex> lock(mutex_);
    while (true) {
      while (!shutdown_ && functions_.empty() && !timer_queue_.timer_fired()) {
        timer_queue_.wait_until_next_deadline(thread_condition_, lock, false);
      }
      if (shutdown_) {
        break;
      }

      std::function<void()> fn;
      // Prioritise timers
      if (timer_queue_.timer_fired()) {
        unsigned int timer_id = timer_queue_.pop();
        std::map<unsigned int, std::function<void()>>::iterator it = timer_functions_.find(timer_id);
        fn = std::move(it->second);
        timer_functions_.erase(it);
        free_timer_ids_.push_back(timer_id);
      } else {
        fn = std::move(functions_.front());
        functions_.pop();
      }

      Unlock unlock(lock);
      fn();
    }
  }
};

/// A simple helper class to assist with creating multi-threaded gmocks.
///
/// This accumulates background actions in a queue and provides an easy way to execute them. The intended use is that
/// you create a single instance of SingleThreadedGMockHelper inside your test case, and use InvokeAsync to invoke
/// actions asynchronously. You should then periodically call coco_drain_queue to execute the accumulated actions.
/// For example:
///
///     ClientMock client_mock;
///     coco::SingleThreadedGMockHelper helper;
///     ...
///     EXPECT_CALL(client_mock, pinger_pong(value))
///       .WillOnce(helper.InvokeAsync(Invoke(&client_mock.pinger().connected(), &PPingerProvided::terminate)))
///
///     while (...) { helper.coco_drain_queue(); }
class SingleThreadedGMockHelper : public SingleThreadedSubscriber, public SingleThreadedTimerExecutor {
 private:
  class ExecuteAsyncAction;

 public:
  SingleThreadedGMockHelper() {}

  ~SingleThreadedGMockHelper() { coco_drain_queue(); }

  /// Executes an arbitrary gMock action asynchronously.
  ///
  /// Note that this will be immediately added to a queue which will execute concurrently with this, meaning that the
  /// execution of the action could happen before this function even returns.
  testing::PolymorphicAction<ExecuteAsyncAction> InvokeAsync(const testing::Action<void()>& action) {
    return testing::MakePolymorphicAction(ExecuteAsyncAction(*this, action, CocoClock::duration::zero()));
  }

  /// Executes an arbitrary gMock action asynchronously after the specified delay has elapsed.
  testing::PolymorphicAction<ExecuteAsyncAction> InvokeAsync(const testing::Action<void()>& action,
                                                             CocoClock::duration delay) {
    return testing::MakePolymorphicAction(ExecuteAsyncAction(*this, action, delay));
  }

  void execute_async(std::function<void()>&& function) { functions_.push(std::move(function)); }

  void execute_async(std::function<void()>&& function, CocoClock::duration delay) {
    unsigned int timer_id = allocate_timer_id();
    timer_queue_.enable_timer(timer_id, delay);
    timer_functions_.insert({timer_id, std::move(function)});
  }

  void coco_drain_queue() COCO_OVERRIDE {
    while (!functions_.empty() || timer_queue_.timer_enabled()) {
      // Wait for any timers to expire
      while (functions_.empty() && !timer_queue_.timer_fired()) {
#ifdef COCO_SLEEP_WORKAROUND
        CocoClock::duration sleep = timer_queue_.next_deadline() - CocoClock::now();
        CocoClock::duration sleep_s = std::chrono::duration_cast<std::chrono::seconds>(sleep);
        CocoClock::duration sleep_ns = std::chrono::duration_cast<std::chrono::nanoseconds>(sleep - sleep_s);
        timespec time_spec;
        time_spec.tv_sec = sleep_s.count();
        time_spec.tv_nsec = sleep_ns.count();
        nanosleep(&time_spec, NULL);
#else
        std::this_thread::sleep_until(timer_queue_.next_deadline());
#endif
      }

      // Prioritise timers
      std::function<void()> fn;
      if (timer_queue_.timer_fired()) {
        unsigned int timer_id = timer_queue_.pop();
        auto it = timer_functions_.find(timer_id);
        fn = std::move(it->second.function);
        if (it->second.component) {
          component_timer_map_.erase(std::make_pair(it->second.component, it->second.component_timer_id));
        }
        timer_functions_.erase(it);
        free_timer_ids_.push_back(timer_id);
      } else {
        fn = std::move(functions_.front());
        functions_.pop();
      }

      fn();
    }
  }

  void enable_timer(BaseSingleThreadedComponent& component, int timer_id, CocoClock::duration duration) COCO_OVERRIDE {
    unsigned int global_timer_id = allocate_timer_id();
    component_timer_map_[std::make_pair(&component, timer_id)] = global_timer_id;
    timer_queue_.enable_timer(global_timer_id, duration);

    PendingTimer& timer = timer_functions_[global_timer_id];
    timer.function = std::bind(&BaseSingleThreadedComponent::execute_timer, &component, timer_id);
    timer.component = &component;
    timer.component_timer_id = timer_id;
  }

  void cancel_timer(BaseSingleThreadedComponent& component, int timer_id) COCO_OVERRIDE {
    auto it = component_timer_map_.find({&component, timer_id});
    if (it != component_timer_map_.end()) {
      timer_functions_.erase(it->second);
      timer_queue_.cancel_timer(it->second);
      free_timer_ids_.push_back(it->second);
      component_timer_map_.erase(it);
    }
  }

 private:
  class ExecuteAsyncAction {
   public:
    ExecuteAsyncAction(SingleThreadedGMockHelper& executor, const testing::Action<void()>& action,
                       CocoClock::duration delay)
        : executor(executor), action(action), delay(delay) {}

    template <typename Result, typename ArgumentTuple>
    Result Perform(const ArgumentTuple& args) {
      if (delay == CocoClock::duration::zero()) {
        executor.execute_async(Executor(action));
      } else {
        executor.execute_async(Executor(action), delay);
      }
    }

   private:
    SingleThreadedGMockHelper& executor;
    testing::Action<void()> action;
    CocoClock::duration delay;

    class Executor {
     public:
      Executor(const testing::Action<void()>& action) : action(action) {}
      void operator()() { action.Perform(testing::Action<void()>::ArgumentTuple()); }

     private:
      testing::Action<void()> action;
    };
  };

  struct PendingTimer {
    PendingTimer() = default;

    PendingTimer(std::function<void()>&& function)
        : function(std::move(function)), component(NULL), component_timer_id(0) {}

    PendingTimer(std::function<void()>&& function, BaseSingleThreadedComponent* component, int component_timer_id)
        : function(std::move(function)), component(component), component_timer_id(component_timer_id) {}

    std::function<void()> function;
    BaseSingleThreadedComponent* component;
    int component_timer_id;
  };

  std::queue<std::function<void()>> functions_;
  std::map<unsigned int, PendingTimer> timer_functions_;
  std::vector<unsigned int> free_timer_ids_;
  std::map<std::pair<BaseSingleThreadedComponent*, int>, unsigned int> component_timer_map_;
  VariableSizeTimerQueue<unsigned int> timer_queue_;

  unsigned int allocate_timer_id() {
    if (free_timer_ids_.empty()) {
      return timer_functions_.size();
    }
    unsigned int timer_id = free_timer_ids_.back();
    free_timer_ids_.pop_back();
    return timer_id;
  }
};

}  // namespace coco

#undef COCO_SLEEP_WORKAROUND

#endif
